/*
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to
 * whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice
 * (including the next paragraph) shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL VIA, S3 GRAPHICS, AND/OR
 * ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "drmP.h"
#include "via_chrome9_drm.h"
#include "via_chrome9_drv.h"
#include "via_chrome9_dma.h"
#include "via_chrome9_mm.h"
#include "via_chrome9_3d_reg.h"
#include "via_chrome9_init.h"
#include "via_chrome9_object.h"
#include "via_chrome9_fence.h"

static int via_chrome9_dp_status(void);
static void via_chrome9_clear_dp_status(void);

static ssize_t dp_hotplug_status_show(struct device *device,
		struct device_attribute *attr,
		char *buf)
{
	int status;
	int size;

	status = via_chrome9_dp_status();
	size = snprintf(buf, PAGE_SIZE, "%d\n", status);
	via_chrome9_clear_dp_status();
	return size;
}

static struct device_attribute dp_hotplug = __ATTR_RO(dp_hotplug_status);
/*fence interrupt function definition*/
void via_chrome9_fence_irq_preinstall(struct drm_via_chrome9_private *dev_priv)
{
	struct drm_device *dev = dev_priv->ddev;
	unsigned long interrupt_status = 0;

	interrupt_status = getmmioregister(dev_priv->mmio_map,INTERRUPT_CONTROL_REG);
	if(!(interrupt_status & INTERRUPT_ENABLE))
		setmmioregister(dev_priv->mmio_map, INTERRUPT_CONTROL_REG,INTERRUPT_ENABLE);
	device_create_file(&dev->primary->kdev, &dp_hotplug);
}
void via_chrome9_fence_irq_uninstall(struct drm_via_chrome9_private *dev_priv)
{
	struct drm_device *dev = dev_priv->ddev;
	unsigned long interrupt_status;	
	interrupt_status = getmmioregister(dev_priv->mmio_map,INTERRUPT_CONTROL_REG);
	/*if other interrupt enable bits are not set, then dsiable hw gfx's intertupt*/
	if(!(interrupt_status & INTERRUPT_ENABLE_MASK))
		setmmioregister(dev_priv->mmio_map,INTERRUPT_CONTROL_REG,
			interrupt_status & (~INTERRUPT_ENABLE));	
	device_remove_file(&dev->primary->kdev, &dp_hotplug);	
}
void via_chrome9_fence_irq_postinstall(struct drm_via_chrome9_private *dev_priv)
{

}

static int dp_status = 0;
static int via_chrome9_dp_status(void)
{
	return dp_status;
}
static void via_chrome9_clear_dp_status(void)
{
	dp_status = 0;
}

irqreturn_t via_chrome9_fence_irq_handler(DRM_IRQ_ARGS)
{
	int interrupt_status2 = 0;
	struct drm_via_chrome9_private *dev_priv = arg;
	u32 mm_1280;
#if 0
	struct list_head *cur_p, *pfence_signaled, *pfence_emited, *tmp;
	struct via_chrome9_fence_object *cur_fence_object = NULL;
	int i;
	unsigned long seq_queried;

#endif

    /**
     * 410MB DP(or internal HDMI), DP-2 interrupt status clean.
     * because our DP(or internal HDMI), DP-2 interrupt enable bit is tied to 1 by HW.
     * so when device hotplug-in, the interrupt will be triggered,
     * so we should handle it here!
	 */
    if ((mm_1280 = getmmioregister(dev_priv->mmio_map, 0x1280 )) & 0x2800 ) {
            setmmioregister(dev_priv->mmio_map, 0x1280, 
					getmmioregister(dev_priv->mmio_map, 0x1280 ) );
					dp_status = mm_1280;
            return IRQ_HANDLED;
    }

	interrupt_status2 = getmmioregister(dev_priv->mmio_map,INTERRUPTCTLREG2);
	if(interrupt_status2 & CR_INT_STATUS)
	{
		setmmioregister(dev_priv->mmio_map, INTERRUPTCTLREG2,
			interrupt_status2 | CR_INT_STATUS);
		/* must read these registers, read fence id high 8 and low 24 bits */
		getmmioregister(dev_priv->mmio_map,0x494);
		getmmioregister(dev_priv->mmio_map,0x49c);
#if 0
		pfence_signaled = &dev_priv->p_fence->signaled;
		spin_lock(&dev_priv->p_fence->fence_pool_lock);

		for (i = 0; i < VIA_FENCE_NUM; i++) {
			tmp = NULL;
			cur_fence_object = NULL;
			pfence_emited = &dev_priv->p_fence->emited[i];
			seq_queried = readl(dev_priv->p_fence->fence_sync_virtual +
						VIA_CHROME9_FENCE_SIZE/4 * i);
			list_for_each(cur_p, pfence_emited) {
				cur_fence_object = container_of(cur_p,
				struct via_chrome9_fence_object, list);
				if (cur_fence_object->sequence <= seq_queried)
					tmp = cur_p;
				else
					break;
			}

			if (cur_fence_object && tmp) {
				cur_p = tmp;
				while (cur_p != pfence_emited) {
					tmp = cur_p->prev;
					list_del_init(cur_p);
					list_add_tail(cur_p, pfence_signaled);
					cur_fence_object  = container_of(cur_p,
						struct via_chrome9_fence_object, list);
					cur_fence_object->signaled = true;
					if (cur_fence_object->waited) {
						wake_up_all(&dev_priv->p_fence->fence_wait_queue);
						cur_fence_object->waited = false;
					}
					cur_p = tmp;
				}
			}
		}

		spin_unlock(&dev_priv->p_fence->fence_pool_lock);
#else
		queue_work(dev_priv->p_fence->fence_wq, &dev_priv->p_fence->fence_work);
#endif
		return IRQ_HANDLED;
	}
	return IRQ_NONE;
}

/*the driver irq function*/
void via_chrome9_driver_irq_preinstall(struct drm_device * dev)
{
	struct drm_via_chrome9_private *dev_priv = 
		(struct drm_via_chrome9_private *) dev->dev_private;
	via_chrome9_fence_irq_preinstall(dev_priv);

}

int via_chrome9_driver_irq_postinstall(struct drm_device *dev)
{
	return 0;
}

void via_chrome9_driver_irq_uninstall(struct drm_device *dev)
{
	struct drm_via_chrome9_private *dev_priv = 
		(struct drm_via_chrome9_private *) dev->dev_private;

	if(drm_core_check_feature(dev_priv->ddev,DRIVER_MODESET))
		via_chrome9_fence_irq_uninstall(dev_priv);
}

irqreturn_t via_chrome9_driver_irq_handler(DRM_IRQ_ARGS)
{
	struct drm_device *dev = arg;
	struct drm_via_chrome9_private *dev_priv = 
		(struct drm_via_chrome9_private *) dev->dev_private;
	return 	via_chrome9_fence_irq_handler(0,dev_priv);
}
